#ifndef READ_DIGEST_H
#define READ_DIGEST_H 1

#include "xml_expat.h"

/**
 * Parses the content of the input configured into the given parser.
 * This is a convenience wrapper for the sequence:
 * <pre>
 * digest_ctx = xoxa_digest_initialize(parser, digest_name);
 * parse_status = xoxa_parse(parser);
 * digest = xoxa_digest_finalize(parser, digest_ctx, digest_len_ptr);
 * return digest;
 * </pre>
 * <p>The list of available digests depends somewhat on the compilation
 * context, but should include for example <code>"SHA1"</code>.
 * <p>If there is an error, return NULL; an error message will be
 * available from xoxa_error_message(ParserState*).
 *
 * @param parser the parser state
 * @param digest_name the name of a digest algorithm
 * @param digest_len a non-null pointer to an integer, which will receive the
 *        length of the returned digest, in bytes
 * @return a pointer to a message digest, or NULL on error
 */
byte* xoxa_parse_for_digest(ParserState* parser,
                            const char* digest_name,
                            unsigned int* digest_len);

/**
 * Initialise the parser for calculating a message digest of the content.
 * The function returns a pointer which must be passed to
 * xoxa_digest_finalize after the parse is complete: this frees
 * memory, and so must be called whether or not the parse was
 * successful.
 * <p>The list of available digests depends somewhat on the compilation
 * context, but should include for example <code>"SHA1"</code>.
 *
 * @param parser the Xoxa parser
 * @param digest_name the name of a digest algorithm
 * @return a pointer which must be passed to xoxa_digest_finalize, or
 *         NULL on error
 */
void* xoxa_digest_initialize(ParserState* parser,
                             const char* digest_name);

/**
 * Finalises the digest calculation.
 * The returned pointer points to allocated memory, which must be
 * freed by the caller after use.
 *
 * @param parser the parser state
 * @param digest_ctx_ptr the return value of a previous
 *        xoxa_digest_initialise function
 * @param digest_len_ptr a non-null pointer to an integer, which will receive the
 *        length of the returned digest, in bytes
 * @return a pointer to a message digest, or NULL on error
 */
byte* xoxa_digest_finalize(ParserState* parser,
                           void* digest_ctx_ptr,
                           unsigned int* digest_len_ptr);


/**
 * Returns some identification of the library providing the digest services.
 * The returned string should not be freed.
 */
const char* xoxa_digest_library_ident(void);

#endif /* READ_DIGEST_H */
