#include <stdio.h>
#include <assert.h>

#include "xml_expat.h"
#define XOXA_PRIVATE_INTERFACE 1
#include "util.h"
#undef XOXA_PRIVATE_INTERFACE
#include "digest.h"

#include "config.h"

// FREE_AND_NULLIFY: free and nullify the argument
#define FREE_AND_NULLIFY(x) if (x != NULL) { \
    free(x);                     \
    x = NULL;                    \
  }


#if HAVE_LIBCRYPTO
// digest_openssl.c needs to be compiled without -Wall on OS X, since from OS X
// 10.9, Apple have marked the entire OpenSSL API as 'deprecated'
// (cf http://rentzsch.tumblr.com/post/33696323211/wherein-i-write-apples-technote-about-openssl-on-os-x;
// on OS X, the module digest_osx.c should be preferred by the configure.ac script).

// Pragmas for clang
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"

// if we HAVE_LIBCRYPTO, we presumably have the headers, too
// (I'm not sure what to do in the alternative case)
#if HAVE_OPENSSL_EVP_H
#include <openssl/evp.h>
#include <openssl/opensslv.h>
#endif

const char* openssl_digest_library_ident(void)
{
    return OPENSSL_VERSION_TEXT;
}

static void openssl_callback(const byte* bytes, size_t len, void* data)
{
    EVP_MD_CTX* mdctx = (EVP_MD_CTX*) data;

    if (! EVP_DigestUpdate(mdctx, bytes, len)) {
        fprintf(stderr, "Can't call digestupdate with bytes %s", bytes);
        exit(1);
    }
}



void* xoxa_digest_initialize(ParserState* parser,
                             const char* digest_name)
{
    EVP_MD_CTX* mdctx;
    const EVP_MD *md_ident;


    OpenSSL_add_all_digests();

    // The manpage isn't clear, but suggests that this return value
    // points to a static value and does not need to be freed.
    md_ident = EVP_get_digestbyname(digest_name);

    if (md_ident == NULL) {
        set_error_message(parser, "digest_openssl: unknown message digest %s", digest_name);
        return NULL;
    }

    if ((mdctx = malloc(sizeof(EVP_MD_CTX))) == NULL) {
        set_error_message(parser, "digest_openssl: can't allocate %ld bytes for md context", sizeof(EVP_MD_CTX));
        return NULL;
    }
    // This mdctx is freed in the _finalize function
    
    EVP_MD_CTX_init(mdctx);
    if (! EVP_DigestInit_ex(mdctx, md_ident, NULL)) {
        set_error_message(parser, "digest_openssl: can't initialise md context for digest name %s", digest_name);
        return NULL;
    }

    xoxa_register_output_callback(parser, &openssl_callback, mdctx);

    assert(mdctx != NULL);
    return (void*)mdctx;
}

byte* xoxa_digest_finalize(ParserState* parser,
                           void* digest_ctx_ptr,
                           unsigned int* digest_len_ptr)
{
    EVP_MD_CTX* mdctx = (EVP_MD_CTX*)digest_ctx_ptr;

    byte* md_value = NULL;
    int all_ok = 1;
    unsigned int digest_length = 0;

    if (digest_len_ptr == NULL) {
        set_error_message(parser, "digest_openssl: xoxa_parse_for_digest requires non-null length argument");
        // don't immediately return NULL, since the cleanup function
        // below may need to release resources
        all_ok = 0;
    }

    if ((md_value = (byte*)malloc(EVP_MAX_MD_SIZE)) == NULL) {
        fatal("Can't allocate %d bytes for digest", EVP_MAX_MD_SIZE);
    }
    // this md_value is freed by the caller

    if (! EVP_DigestFinal_ex(mdctx, md_value, &digest_length)) {
        set_error_message(parser, "digest_openssl: can't finalise md context");
        all_ok = 0;
    }

    EVP_MD_CTX_cleanup(mdctx);
    FREE_AND_NULLIFY(mdctx);

    if (all_ok) {
        assert(digest_len_ptr != NULL);
        assert(digest_length > 0); // sanity check
        *digest_len_ptr = digest_length;
    } else {
        FREE_AND_NULLIFY(md_value);
    }

    assert((all_ok && (md_value != NULL))
           || (!all_ok && (md_value == NULL)));
    return md_value;
}

static byte* parse_for_digest_openssl(ParserState* parser,
                                      const char* digest_name,
                                      unsigned int* digest_len_ptr)
{
    void* digest_ctx;
    byte* digest;
    int parse_status;

    assert(digest_len_ptr != NULL);

    digest_ctx = xoxa_digest_initialize(parser, digest_name);
    parse_status = xoxa_parse(parser);
    digest = xoxa_digest_finalize(parser, digest_ctx, digest_len_ptr);

    if (parse_status != 0) {
        FREE_AND_NULLIFY(digest);
    }

    assert((parse_status != 0 && digest == NULL)
           || (parse_status == 0 && digest != NULL));

    return digest;
}

#pragma clang diagnostic pop
#endif /* HAVE_LIBCRYPTO */

byte* xoxa_parse_for_digest(ParserState* parser,
                       const char* digest_name,
                       unsigned int* digest_len_ptr)
{
#if HAVE_LIBCRYPTO
    byte* digest;

    if (parser == NULL) {
        fprintf(stderr, "Bad call: xoxa_parse_for_digest requires initialised parser for digest %s\n", digest_name);
        return NULL;
    }
    if (digest_len_ptr == NULL) {
        fprintf(stderr, "Bad call: xoxa_parse_for_digest non-null digest_len_ptr\n");
        return NULL;
    }

    digest = parse_for_digest_openssl(parser, digest_name, digest_len_ptr);

    return digest;
#else
    return NULL;
#endif
}

const char* xoxa_digest_library_ident(void)
{
#if HAVE_LIBCRYPTO
    return openssl_digest_library_ident();
#else
    return NULL;
#endif
}
