#include <CommonCrypto/CommonDigest.h>
#include <string.h> // for memcpy
#include <ctype.h> // for tolower
#include <assert.h>

#define XOXA_PRIVATE_INTERFACE 1
#include "util.h"
#undef XOXA_PRIVATE_INTERFACE

#include "digest.h"

// FREE_AND_NULLIFY: free and nullify the argument
#define FREE_AND_NULLIFY(x) if (x != NULL) { \
    free(x);                     \
    x = NULL;                    \
  }
    

static void sha1_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA1_Update((CC_SHA1_CTX*)data, (void*)bytes, len);
}

static void sha224_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA224_Update((CC_SHA256_CTX*)data, (void*)bytes, len);
}

static void sha256_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA256_Update((CC_SHA256_CTX*)data, (void*)bytes, len);
}

static void sha384_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA384_Update((CC_SHA512_CTX*)data, (void*)bytes, len);
}

static void sha512_callback(const byte* bytes, size_t len, void* data)
{
    CC_SHA512_Update((CC_SHA512_CTX*)data, (void*)bytes, len);
}

static void md5_callback(const byte* bytes, size_t len, void* data)
{
    CC_MD5_Update((CC_MD5_CTX*)data, (void*)bytes, len);
}

enum digests { invalid, SHA1, SHA2_224, SHA2_256, SHA2_384, SHA2_512, MD5 };
static struct {
    const char* name;
    const enum digests ident;
} all_digests[] = {
    { "md5", MD5 },
    { "sha1", SHA1 },
    { "sha224", SHA2_224 },
    { "sha256", SHA2_256 },
    { "sha384", SHA2_384 },
    { "sha512", SHA2_512 },
};
int n_all_digests = sizeof(all_digests)/sizeof(all_digests[0]);

// The following type matches each of the actual finalizer functions,
// with the second argument cast to void*.
// This lets us avoid having a switch on the digest_ident in both the
// initializer and finalizer function.  This admittedly adds a bit of
// trickiness (including for example this cast), but means they can't
// be out of sync.
typedef int (*cc_generic_finalizer_function)(unsigned char *md, void *c);
typedef struct {
    void* ctx;
    enum digests digest_ident;
    unsigned int digest_length;
    cc_generic_finalizer_function finalizer;
} xoxa_digest_ctx;


void* xoxa_digest_initialize(ParserState* parser,
                             const char* digest_name)
{
    xoxa_digest_ctx* digest_ctx;
    enum digests digest_ident = invalid;


    if (parser == NULL) {
        // can't set this to parser->error_message !
        fprintf(stderr, "digest_osx: xoxa_digest_initialize requires initialised parser for digest %s\n", digest_name);
        return NULL;
    }

    {
        int i;
        for (i=0; i<n_all_digests; i++) {
            if (strcasecmp(digest_name, all_digests[i].name) == 0) {
                digest_ident = all_digests[i].ident;
                break;
            }
        }
        if (i == n_all_digests) {
            set_error_message(parser, "digest_osx: unrecognised digest: %s", digest_name);
            return NULL;
        }
    }
    assert(digest_ident != invalid);

#define DO_DIGEST(ctxtype, digestname, callback, length)                \
    {                                                                   \
        if ((digest_ctx->ctx = malloc(sizeof(ctxtype))) == NULL) {      \
            fatal("digest_osx: unable to allocate %d bytes for message context", sizeof(ctxtype)); \
        }                                                               \
        digest_ctx->digest_length = length;                             \
        digest_ctx->finalizer = (cc_generic_finalizer_function)&digestname ## _Final; \
        digestname ## _Init(digest_ctx->ctx);                           \
        xoxa_register_output_callback(parser, &callback, digest_ctx->ctx); \
    }    

    if ((digest_ctx = malloc(sizeof(xoxa_digest_ctx))) == NULL) {
        fatal("digest_osx: unable to allocate %d bytes for message context",
              sizeof(xoxa_digest_ctx));
    }
    digest_ctx->digest_ident = digest_ident;

    switch (digest_ident) {
      case SHA1:
        DO_DIGEST(CC_SHA1_CTX, CC_SHA1, sha1_callback, CC_SHA1_DIGEST_LENGTH);
        break;
      case SHA2_224:
        DO_DIGEST(CC_SHA256_CTX, CC_SHA224, sha224_callback, CC_SHA224_DIGEST_LENGTH);
        break;
      case SHA2_256:
        DO_DIGEST(CC_SHA256_CTX, CC_SHA256, sha256_callback, CC_SHA256_DIGEST_LENGTH);
        break;
      case SHA2_384:
        DO_DIGEST(CC_SHA512_CTX, CC_SHA384, sha384_callback, CC_SHA384_DIGEST_LENGTH);
        break;
      case SHA2_512:
        DO_DIGEST(CC_SHA512_CTX, CC_SHA512, sha512_callback, CC_SHA512_DIGEST_LENGTH);
        break;
      case MD5:
        DO_DIGEST(CC_MD5_CTX, CC_MD5, md5_callback, CC_MD5_DIGEST_LENGTH);
        break;
      default:
        assert(0);
    }
#undef DO_DIGEST

    return (void*)digest_ctx;
}

byte* xoxa_digest_finalize(ParserState* parser,
                           void* digest_ctx_ptr,
                           unsigned int* digest_len_ptr)
{
    xoxa_digest_ctx* digest_ctx = (xoxa_digest_ctx*)digest_ctx_ptr;
    byte* md = NULL;
    unsigned int digest_length = 0;
    int all_ok = 1;

    
    if (digest_len_ptr == NULL) {
        set_error_message(parser, "digest_osx: xoxa_digest_finalize requires non-null length argument");
        // don't simply return NULL here, because we do want to call
        // the CC finalizer function, in case that releases resources
        all_ok = 0;
    }

    digest_length = digest_ctx->digest_length;

    if ((md = malloc(digest_length)) == NULL) {
        fatal("digest_osx: unable to allocate %d bytes for message digest", digest_length);
    }
    assert(md != NULL);

    (*digest_ctx->finalizer)(md, (void*)digest_ctx->ctx);
    
    FREE_AND_NULLIFY(digest_ctx);

    if (all_ok) {
        assert(digest_len_ptr != NULL);
        *digest_len_ptr = digest_length;
    } else {
        // bad-call error earlier
        FREE_AND_NULLIFY(md);
    }

    assert((all_ok && (md != NULL))
           || (!all_ok && (md == NULL)));

    return md;
}

byte* xoxa_parse_for_digest(ParserState* parser,
                            const char* digest_name,
                            unsigned int* digest_len_ptr)
{
    void* digest_ctx;
    byte* digest;
    int parse_status;

    if (digest_len_ptr == NULL) {
        set_error_message(parser, "digest_osx: xoxa_parse_for_digest requires non-null length argument");
        return NULL;
    }

    digest_ctx = xoxa_digest_initialize(parser, digest_name);
    parse_status = xoxa_parse(parser);
    digest = xoxa_digest_finalize(parser, digest_ctx, digest_len_ptr);

    if (parse_status != 0 || digest == NULL)  {
        // failure
        if (digest != NULL) {
            FREE_AND_NULLIFY(digest);
        }
        assert(digest == NULL);
    } else {
        // success
        assert(digest != NULL);
    }

    return digest;
}

#include <sys/sysctl.h>
const char* xoxa_digest_library_ident(void)
{
    static char res[256] = "";

    if (res[0] == '\0') {
        // first time
        char version[256];
        size_t version_size = sizeof(version);
        Buf* digest_list;
        int i;

        (void) sysctlbyname("kern.osrelease", version, &version_size, NULL, 0);

        digest_list = bufinit();

        for (i=0; i<n_all_digests; i++) {
            bufprintf(digest_list, " %s", all_digests[i].name);
        }
        snprintf(res, sizeof(res),
                 "OS X Common Crypto (%s); available digests:%s",
                 version, bufstring(digest_list));

        buffree(digest_list);
    }
    return res;
}
