#define XOXA_PRIVATE_INTERFACE 1
#include "xml_expat.h"
#include "util.h"
#undef XOXA_PRIVATE_INTERFACE

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <ctype.h>

#define ALLOC(var, Type) if ((var = (Type*)malloc(sizeof(Type))) == NULL) { \
        fatal("Can't allocate space for " # Type);                      \
    }

static int output_string_with_whitespace_skipping(Buf* buf,
                                                  const XML_Char* c,
                                                  size_t len,
                                                  char *skipping_whitespace);
static void write_out_accumulated_text(ParserState* P);
static ParserState* make_new_parserstate(InputSource* src, XML_Parser parser);
static int xoxa_parse_file(ParserState* P);

// flags for ParserState->text_skipping_whitespace
#define SKIPPING_WHITESPACE_COLLAPSE 0
#define SKIPPING_WHITESPACE_TRUE 1
#define SKIPPING_WHITESPACE_NONORM 2



/**
 * Creates an InputSource from the named file.
 * The InputSource should be handed to a {@link xoxa_new_parser} function,
 * which will free it when necessary.
 * Returns NULL if there's any problem.
 * (in fact, the only problem is memory exhaustion, which causes an
 * immediate exit)
 * @param filename the file which is to be read in
 * @return a new InputSource
 */
InputSource* xoxa_new_input_source_from_filename(const char* filename)
{
    InputSource* src;
    FILE* file;

    ALLOC(src, InputSource);

    if ((file = fopen(filename, "r")) == NULL) {
        src = NULL;
    } else {
        src->type = INPUT_SOURCE_FILE;
        src->basename = filename;
        src->in.f = file;
    }

    return src;
}

/**
 * Creates an InputSource from the named file descriptor.
 * The InputSource should be handed to a {@link xoxa_new_parser} function,
 * which will free it when necessary.
 * Returns NULL if there's any problem.
 * (in fact, the only problem is memory exhaustion, which causes an
 * immediate exit)
 * @param fd the identifier of the file descriptor which is to be read in
 * @return a new InputSource
 */
InputSource* xoxa_new_input_source_from_fd(const int fd)
{
    InputSource* src;

    ALLOC(src, InputSource);

    src->type = INPUT_SOURCE_FILE;
    src->basename = NULL;
    src->in.f = fdopen(fd, "r");

    return src;
}    

/**
 * Creates an InputSource from the given content, which should be an
 * XML string.  This string is not copied, so should not be freed
 * during a parse.
 * The InputSource should be handed to a {@link xoxa_new_parser} function,
 * which will free it when necessary.
 * Returns NULL if there's any problem.
 * (in fact, the only problem is memory exhaustion, which causes an
 * immediate exit)
 * @param xmlcontent the string which is to be read in
 * @return a new InputSource
 */
InputSource* xoxa_new_input_source_from_string(const char* xmlcontent)
{
    InputSource* src;

    ALLOC(src, InputSource);

    src->type = INPUT_SOURCE_STRING;
    src->basename = NULL;
    src->in.buffer = xmlcontent;

    return src;
}

// Separator of namespace and localname used by Expat
#define NS_SEPARATOR '\xFF'

// comparison function for 0-terminated byte strings
static int bstrcmp(const void* p1, const void* p2) 
{
    const byte* b1 = *((byte**)p1);
    const byte* b2 = *((byte**)p2);

    while (1) {
        if (*b1 == 0) {
            if (*b2 == 0) return 0; // equal
            return -1;
        } else if (*b2 == 0) {
            return +1;
        } else if (*b1 == *b2) {
            b1++;
            b2++;
        } else if (*b1 > *b2) {
            return +1;
        } else {
            return -1;
        }
    }
}

// Expat handler functions

static void expat_element_start(void *data, const XML_Char *el, const XML_Char **attr)
{
    int i;
    ParserState* P = (ParserState*)data;
    char* nsp;
    byte* output;
#define MAXATT 32
    // make this statically allocated for the moment
    // (not pretty, but 32 is a lot of attributes...)
    size_t fmtatt[MAXATT];
    byte* fmtatt_b[MAXATT];
    int natt;
    char skipping_whitespace_init;

    assert(P != NULL);
    assert(P->callback != NULL);
    assert(P->B != NULL);

    // if there are other callbacks, call them now
    if (P->extra_StartElementHandler) {
        (*P->extra_StartElementHandler)(P->extra_UserData, el, attr);
    }

    write_out_accumulated_text(P);

    bufreset(P->B);

    if (P->parser_flags & XOXA_REPORT_ALL) {
        skipping_whitespace_init = SKIPPING_WHITESPACE_NONORM;
    } else {
        skipping_whitespace_init = SKIPPING_WHITESPACE_COLLAPSE;
    }

    // build up an array of attribute strings
    natt = 0;
    for (i=0; attr[i]; i+=2) {
        // skipping_whitespace is potentially modified in output_string_with_whitespace_skipping
        char skipping_whitespace;

        if (natt == MAXATT) {
            fatal("Unexpectedly large number (>%d) of attributes on element %s", MAXATT, el);
        }

        fmtatt[natt] = buflength(P->B); // use this as an offset into the buffer
        nsp = strchr(attr[i], NS_SEPARATOR);

        skipping_whitespace = skipping_whitespace_init;

        if (nsp == NULL) {
            bufprintf(P->B, "A%s CDATA ", attr[i]);
            output_string_with_whitespace_skipping(P->B,
                                                   attr[i+1],
                                                   0,
                                                   &skipping_whitespace);
            bufprintf(P->B, "%c", NS_SEPARATOR);
        } else {
            bufprintf(P->B, "B%.*s %s CDATA ", nsp-attr[i], attr[i], nsp+1);
            output_string_with_whitespace_skipping(P->B,
                                                   attr[i+1],
                                                   0,
                                                   &skipping_whitespace);
            bufprintf(P->B, "%c", NS_SEPARATOR);
        }
        natt++;
    }

    // sort these...
    output = bufstring_copy(P->B);
    // first, split this output at NS_SEPARATOR
    for (i=0; i<natt; i++) {
        byte* p;
        fmtatt_b[i] = output+fmtatt[i];
        for (p=fmtatt_b[i]; *p!=(byte)NS_SEPARATOR; p++) {
            // nothing
        }
        *p = '\0';
    }
    qsort((void*)fmtatt_b, natt, sizeof(fmtatt_b[0]), &bstrcmp);

    for (i=0; i<natt; i++) {
        byte crlf[2] = {'\r', '\n'};
        
        P->callback(fmtatt_b[i], strlen((const char*)fmtatt_b[i]), P->callback_data);
        P->callback(crlf, 2, P->callback_data);
    }
    free(output);

    bufreset(P->B);

    nsp = strchr(el, NS_SEPARATOR);
    if (nsp == NULL) {
        // no namespace
        P->B = bufprintf(P->B, "(%s\r\n", el);
    } else {
        P->B = bufprintf(P->B, "[%.*s %s\r\n", nsp-el, el, nsp+1);
    }

    P->callback(bufstring(P->B), buflength(P->B), P->callback_data);

    // Set the parser's skipping-whitespace flag to the correct
    // initial value for processing the text content of the element.
    P->text_skipping_whitespace = skipping_whitespace_init;
#undef MAXATT
}

static void expat_element_end(void *data, const XML_Char *el)
{
    ParserState* P = (ParserState*)data;
    char* nsp;

    assert(P != NULL);
    assert(P->callback != NULL);
    assert(P->B != NULL);

    if (P->extra_EndElementHandler) {
        (*P->extra_EndElementHandler)(P->extra_UserData, el);
    }

    write_out_accumulated_text(P);

    bufreset(P->B);

    nsp = strchr(el, NS_SEPARATOR);
    if (nsp == NULL) {
        // no namespace
        P->B = bufprintf(P->B, ")%s\r\n", el);
    } else {
        P->B = bufprintf(P->B, "]%.*s %s\r\n", nsp-el, el, nsp+1);
    }

    P->callback(bufstring(P->B), buflength(P->B), P->callback_data);
}

static void expat_character_data_handler(void *data,
                                         const XML_Char *s,
                                         int len)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);

    output_string_with_whitespace_skipping(P->text,
                                           s,
                                           len,
                                           &P->text_skipping_whitespace);
}

static void write_out_accumulated_text(ParserState* P)
{
    size_t len;

    assert(P != NULL);
    len = buflength(P->text);

    // if len==0, there's nothing to write;
    // if (len==1 and skipping_whitespace = TRUE),
    //   then all there is to write is whitespace
    //   (and we're in the normalisation mode);
    // in either case do nothing
    if (! (len == 0
           || (len == 1
               && P->text_skipping_whitespace == SKIPPING_WHITESPACE_TRUE))) {
        byte dash[] = { '-' };
        byte crlf[] = { '\r', '\n' };

        P->callback(dash, 1, P->callback_data);
        P->callback(bufstring(P->text), buflength(P->text), P->callback_data);
        P->callback(crlf, 2, P->callback_data);
    }

    if (P->extra_CharacterDataHandler) {
        // XXX: I'm somewhat uncertain about this cast.  There's a check,
        // in xoxa_new_parser, that the sizeof values for byte and
        // XML_Char are the same, but are there sign and
        // unicode-related problems possible, too?
        (*P->extra_CharacterDataHandler)(P->extra_UserData,
                                         (const XML_Char*)bufstring(P->text),
                                         buflength(P->text));
    }

    bufreset(P->text);

    // Set text_skipping_whitespace to the correct value to process
    // the text content correctly.
    if (P->parser_flags & XOXA_REPORT_ALL) {
        P->text_skipping_whitespace = SKIPPING_WHITESPACE_NONORM;
    } else {
        P->text_skipping_whitespace = SKIPPING_WHITESPACE_COLLAPSE;
    }
}

static void expat_namespace_start(void *data, 
                                  const XML_Char *prefix,
                                  const XML_Char *uri)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);
    write_out_accumulated_text(P);

    if (P->extra_StartNamespaceDeclHandler) {
        (*P->extra_StartNamespaceDeclHandler)(P->extra_UserData, prefix, uri);
    }

    if (P->parser_flags & XOXA_REPORT_ALL) {
        bufreset(P->B);
        bufprintf(P->B, "M%s %s\r\n", (prefix ? prefix : ""), uri);
        P->callback(bufstring(P->B), buflength(P->B), P->callback_data);
    }
}

static void expat_namespace_end(void *data, 
                                const XML_Char *prefix)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);
    write_out_accumulated_text(P);

    if (P->extra_StartNamespaceDeclHandler) {
        (*P->extra_EndNamespaceDeclHandler)(P->extra_UserData, prefix);
    }

    if (P->parser_flags & XOXA_REPORT_ALL) {
        bufreset(P->B);
        bufprintf(P->B, "m%s\r\n", (prefix ? prefix : ""));
        P->callback(bufstring(P->B), buflength(P->B), P->callback_data);
    }
}

static void expat_pi(void *data,
                     const XML_Char *target,
                     const XML_Char *pi_data)
{
    ParserState* P = (ParserState*)data;
    int show_it = 1;            // do we write this PI out?


    assert(P != NULL);

    if (strcmp(target, "signature") == 0) {
        if (P->parser_flags & XOXA_REPORT_ALL) {
            show_it = 1;
        } else {
            show_it = 0;
        }

        if (P->signature_fd > 0) {
            // Write the signature block out to the given FD.
            // At present, we concern ourselves only with the GPG
            // signature PIs which have an 'armor="..." 'attribute'.
            const char *armor;
            armor = strstr((const char*)pi_data, "armor");
            int verbose = SHOWING_MESSAGES(P);

            if (armor == NULL) {
                if (verbose) {
                    fprintf(stderr, "No signature block found\n");
                }
            } else {
                const char *p;
                p = strchr(armor, '"');
                if (p == NULL) {
                    if (verbose) {
                        fprintf(stderr, "Signature block found, but armor=\"...\" malformed! <%s>\n", pi_data);
                    }
                } else {
                    const char* p_end;
                    p++;
                    p_end = strchr(p, '"');
                    if (p_end == NULL) {
                        if (verbose) {
                            fprintf(stderr, "Signature block found, but armor=\"...\" malformed! <%s>\n", pi_data);
                        }
                        // ooops, but don't fail, and simply find the
                        // end of the string
                        p_end = strchr(p, '\0');
                    }
                    write(P->signature_fd, p, p_end-p);
                }
            }
        }            
    }

    if (show_it) {
        char skipping_whitespace = SKIPPING_WHITESPACE_COLLAPSE;
        int found_chars;

        write_out_accumulated_text(P);

        bufreset(P->B);
        bufprintf(P->B, "?%s ", target);

        found_chars = output_string_with_whitespace_skipping(P->B,
                                                             pi_data,
                                                             0,
                                                             &skipping_whitespace);
        bufcrop(P->B);
        if (! found_chars) {
            // the preceding crop will have removed _all_ trailing
            // characters, including the one following the "?%s ".
            // So put it back.
            bufchar(P->B, ' ');
        }

        bufprintf(P->B, "\r\n");
        P->callback(bufstring(P->B), buflength(P->B), P->callback_data);
    }

    if (P->extra_ProcessingInstructionHandler && strcmp(target, "signature") != 0) {
        (*P->extra_ProcessingInstructionHandler)(P->extra_UserData, target, pi_data);
    }
}

static void expat_xmldecl(void *data,
                          const XML_Char* version,
                          const XML_Char* encoding,
                          int standalone)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);

    if (strcmp(version, "1.0") != 0) {
        if (SHOWING_MESSAGES(P)) {
            fprintf(stderr, "warning: document is <?xml version='%s'?>, but xoxa/expat can in principle handle only XML 1.0; there may be subtle whitespace normalisation errors.\n", version);
            // The errors in question are to do with handling NEL
            // (U+0085) and line-separator (U+2028); cf XML 1.1 Sect. 1.3.
        }
    }

    if (P->extra_XmlDeclHandler) {
        (*P->extra_XmlDeclHandler)(P->extra_UserData, version, encoding, standalone);
    }
}

static int expat_external_entity_ref_handler(XML_Parser main_parser_state_ptr,
                                             const XML_Char *context,
                                             const XML_Char *document_base_filename,
                                             const XML_Char *systemId,
                                             const XML_Char *publicId)
{
    ParserState* PS = (ParserState*)main_parser_state_ptr;
    XML_Parser main_parser = PS->parser;
    InputSource* entity_src;
    int old_cwd = -1;
    const char* bases[2];
    int base_idx;
    int error_p = 0;

    // We _don't_ call any extra ExternalEntityRefHandler here, but
    // simply arrange that the original callbacks are called as
    // usual.  This is so that there's no opportunity for the caller
    // which lodged such a callback to read the entity.  We want to
    // read it, and we don't want the caller to get in the way.

    if (systemId == NULL) {
        // I think this should be impossible, but it's not a Xoxa
        // error, so fail with an error message
        fatal("Unexpected external entity with empty system id at line %lu",
              XML_GetCurrentLineNumber(main_parser));
    }

    // Assume, at least for the present, that the systemId and base
    // point to a file, rather than a general URI.
    bases[1] = document_base_filename;
    bases[0] = PS->src->basename; // filename of the current document
    base_idx = 1;                 // index of last element of bases[]

    while (base_idx >= 0) {
        if (bases[base_idx] == NULL) {
            base_idx--;         // go round again
        } else {
            const char* base = bases[base_idx];
            const char *slash = strrchr(base, '/');
            if (slash == NULL) {
                // the base URI is just a file -- nothing to do
            } else {
                // we've found the last slash in the string, so this
                // indicates the directory of the base file
                size_t dirlen = slash - base;
                int chdir_status;
                char* new_cwd = malloc(dirlen + 1);

                if (new_cwd == NULL) {
                    fatal("Can't allocate space for base-URI directory, %s (%lu bytes)",
                          bases, dirlen);
                }
                memcpy(new_cwd, base, dirlen);
                new_cwd[dirlen] = '\0';

                // open the current working directory, and save its (nonzero) FD
                old_cwd = open(".", O_RDONLY);

                chdir_status = chdir(new_cwd);

                if (chdir_status != 0) {
                    // ooops
                    set_error_message(PS, "Can't cwd to directory %s (base=%s)", new_cwd, base);
                    free(new_cwd);
                    goto clean_up; // JUMP OUT
                }
                free(new_cwd);
            }
            base_idx = -1;      // exit loop
        }
    }

    //fprintf(stderr, "external entity: %s\n", systemId);

    entity_src = xoxa_new_input_source_from_filename(systemId);
    if (entity_src == NULL) {
        char* cwd = getcwd(NULL, 0);
        set_error_message(PS, "Can't open entity %s to read (cwd=%s)", systemId, cwd);
        free(cwd);
        goto clean_up;
    }

    //fprintf(stderr, "...being parsed...\n");

    {
        XML_Parser sub_parser;
        ParserState* sub_PS;
        int sub_error_p = 0;
    
        sub_parser = XML_ExternalEntityParserCreate(main_parser, context, NULL);
        sub_PS = make_new_parserstate(entity_src, sub_parser);
        XML_SetExternalEntityRefHandlerArg(sub_parser, sub_PS);

        sub_PS->extra_StartElementHandler = PS->extra_StartElementHandler;
        sub_PS->extra_EndElementHandler = PS->extra_EndElementHandler;
        sub_PS->extra_CharacterDataHandler = PS->extra_CharacterDataHandler;
        sub_PS->extra_ProcessingInstructionHandler = PS->extra_ProcessingInstructionHandler;
        sub_PS->extra_StartCdataSectionHandler = PS->extra_StartCdataSectionHandler;
        sub_PS->extra_EndCdataSectionHandler = PS->extra_EndCdataSectionHandler;
        sub_PS->extra_StartNamespaceDeclHandler = PS->extra_StartNamespaceDeclHandler;
        sub_PS->extra_EndNamespaceDeclHandler = PS->extra_EndNamespaceDeclHandler;
        sub_PS->extra_XmlDeclHandler = PS->extra_XmlDeclHandler;
        sub_PS->extra_SkippedEntityHandler = PS->extra_SkippedEntityHandler;
        sub_PS->extra_UserData = PS->extra_UserData;

        sub_error_p = xoxa_parse_file(sub_PS);
        if (sub_error_p) {
            set_error_message(PS, xoxa_error_message(sub_PS));
            error_p = 1;
        }
       
        xoxa_release_parser(sub_PS);     // also frees sub_parser and entity_src
    }

    //fprintf(stderr, "...entity %s parsed\n", systemId);

 clean_up:
    if (old_cwd >= 0) {
        // change back to the original WD
        fchdir(old_cwd);
        close(old_cwd);
    }

    if (error_p) {
        if (SHOWING_MESSAGES(PS)) {
            fprintf(stderr, "Error parsing entity %s: %s\n", systemId, xoxa_error_message(PS));
        }
        return XML_STATUS_ERROR;
    } else {
        return XML_STATUS_OK;
    }

#undef ERR_MSG_LEN
}

static void expat_skipped_entity_handler(void *data,
                                         const XML_Char *entityName,
                                         int is_parameter_entity)
{
    ParserState* P = (ParserState*)data;

    assert(P != NULL);

    if (SHOWING_MESSAGES(P)) {
        fprintf(stderr, "Skipped entity found (%s%s) -- I didn't expect that\n",
                (is_parameter_entity ? "parameter entity: " : ""),
                entityName);
    }

    if (P->extra_SkippedEntityHandler) {
        (*P->extra_SkippedEntityHandler)(P->extra_UserData, entityName, is_parameter_entity);
    }
}

// The default (Xoxa) callback simply writes the byte buffer to stdout.
// If 'len' is zero, then the buf is null-terminated; if not, then it
// gives the number of bytes to be written out.
static void default_callback(const byte* buf, size_t len, void* data)
{
    if (len == 0) {
        for (len = 0; buf[len] != '\0'; len++) {
            // nothing
        }
    }
    fwrite(buf, len, 1, stdout);
}

// API functions

static ParserState* make_new_parserstate(InputSource* src, XML_Parser parser)
{
    ParserState* P;

    ALLOC(P, ParserState);

    P->parser = parser;
    P->src = src;
    P->callback = NULL;
    P->callback_data = NULL;
    P->B = bufinit();
    P->text = bufinit();
    P->text_skipping_whitespace = SKIPPING_WHITESPACE_COLLAPSE;
    P->parser_flags = 0;
    P->signature_fd = 0;
    P->error_message = NULL;

    P->extra_StartElementHandler = NULL;
    P->extra_EndElementHandler = NULL;
    P->extra_CharacterDataHandler = NULL;
    P->extra_ProcessingInstructionHandler = NULL;
    P->extra_StartCdataSectionHandler = NULL;
    P->extra_EndCdataSectionHandler = NULL;
    P->extra_StartNamespaceDeclHandler = NULL;
    P->extra_EndNamespaceDeclHandler = NULL;
    P->extra_XmlDeclHandler = NULL;
    P->extra_SkippedEntityHandler = NULL;
    P->extra_UserData = NULL;

    return P;
}

/**
 * Returns a freshly-initialised parser.
 * After this call, the InputSource object is owned by the Parser, and
 * will be released in xoxa_release_parser.
 * @param src an initialised InputSource object
 * @return a fresh parser
 */
ParserState* xoxa_new_parser(InputSource* src)
{
    ParserState* PS;

    // The following should always be true, unless the Expat library
    // was compiled with XML_UNICODE=1 _and_ the resulting library was
    // not renamed to libexpatw.la as the Expat build instructions
    // require.
#define BUILD_BUG_ON(condition) ((void)sizeof(char[1 - 2*!!(condition)]))
    BUILD_BUG_ON(sizeof(XML_Char) != 1);

    XML_Parser p = XML_ParserCreateNS(NULL, NS_SEPARATOR);
    if (! p) {
        fatal("Couldn't allocate memory for expat parser");
    }

    PS = make_new_parserstate(src, p);
    assert(PS != NULL);

    xoxa_register_output_callback(PS, &default_callback, NULL);

    XML_SetElementHandler(p, expat_element_start, expat_element_end);
    XML_SetCharacterDataHandler(p, expat_character_data_handler);
    XML_SetNamespaceDeclHandler(p,
                                expat_namespace_start,
                                expat_namespace_end);
    XML_SetProcessingInstructionHandler(p, expat_pi);
    XML_SetXmlDeclHandler(p, expat_xmldecl);
    XML_SetExternalEntityRefHandler(p, expat_external_entity_ref_handler);
    XML_SetExternalEntityRefHandlerArg(p, (void*)PS);
    XML_SetSkippedEntityHandler(p, expat_skipped_entity_handler);
    XML_SetUserData(p, (void*)PS);

    return PS;
}

/**
 * Sets parser flags.
 * @param P an initialised parser
 * @param PF one or more flags, or-ed together
 * @return previous flags.
 */
ParserFlags xoxa_parser_flags(ParserState* P, ParserFlags PF)
{
    ParserFlags prev = 0;

    if (P) {
        prev = P->parser_flags;
        P->parser_flags = PF;
    }
    return prev;
}

/**
 * Specify the file-descriptor to which the library should write a signature block.
 * XXX details.
 */
void xoxa_signature_fd(ParserState* P, int fd)
{
    if (P) {
        if (fd < 1) {
            if (SHOWING_MESSAGES(P)) {
                fprintf(stderr, "xoxa_signature_fd: fd (%d) must be greater than 1 -- ignored\n", fd);
            }
        } else {
            P->signature_fd = fd;
        }
    }
}

/**
 * Releases the parser and the contained input source.
 */
void xoxa_release_parser(ParserState* P)
{
    if (P != NULL) {
        assert(P->parser != NULL);
        XML_ParserFree(P->parser);
        P->parser = NULL;
        if (P->src != NULL) {
            free(P->src);
            P->src = NULL;
        }
        if (P->error_message != NULL) {
            free(P->error_message);
            P->error_message = NULL;
        }
        free(P);
    }
}

static int xoxa_parse_string(ParserState* P)
{
    enum XML_Status status;
    int rval;

    assert(P != NULL && P->src->type == INPUT_SOURCE_STRING);

    status = XML_Parse(P->parser,
                       P->src->in.buffer,
                       strlen(P->src->in.buffer),
                       1);
 
   if (status == XML_STATUS_OK) {
       rval = 0;
    } else {
       set_error_message(P, "Parse error at character %lu: %s\n",
                         XML_GetCurrentByteIndex(P->parser),
                         XML_ErrorString(XML_GetErrorCode(P->parser)));
       rval = 1;
   }

   return rval;
}

static int xoxa_parse_file(ParserState* P)
{
    int keep_going = 1;
    FILE* f;
    int rval = 0;

    assert(P != NULL && P->src != NULL && P->src->type == INPUT_SOURCE_FILE);
    assert(P->src->in.f != NULL);

    f = P->src->in.f;

    while (keep_going) {
        int bytes_read;
        int is_last;

        void *buf = XML_GetBuffer(P->parser, BUFSIZ);
        if (buf == NULL) {
            fatal("Unable to allocate memory in expat GetBuffer");
        }

        bytes_read = fread(buf, 1, BUFSIZ, f);
        if (ferror(f)) {
            fatal("Error reading input XML");
        }

        if (feof(f)) {
            keep_going = 0;
            is_last = 1;
        } else {
            is_last = 0;
        }

        if (XML_ParseBuffer(P->parser, bytes_read, is_last) == XML_STATUS_ERROR) {
            const char* fn = P->src->basename;
            set_error_message(P,
                              "Parse error at %s:%lu: %s\n",
                              (fn == NULL ? "<input>" : fn),
                              XML_GetCurrentLineNumber(P->parser),
                              XML_ErrorString(XML_GetErrorCode(P->parser)));
            rval = 1;
        }
    }

    fclose(f);
    P->src->in.f = NULL;

    return rval;
}

/**
 * Parse the input, using the states set up in the given lexer and parser.
 * @param P a configured parser
 * @return NULL on success, and an error message otherwise.
 */
int xoxa_parse(ParserState* P)
{
    int rval;

    if (P == NULL) {
        fprintf(stderr, "Bad call: xoxa_parse received null ParserState\n");
        return 1;
    }
    if (P->src == NULL) {
        set_error_message(P, "Bad call: xoxa_parse has no input source");
        return 1;
    }

    switch (P->src->type) {
      case INPUT_SOURCE_STRING:
        rval = xoxa_parse_string(P);
        break;
      case INPUT_SOURCE_FILE:
        rval = xoxa_parse_file(P);
        break;
      default:
        assert(0);
    }
    return rval;
}

/**
 * Retrieve any error message which has been collected during a parse.
 * @param P a parser, which has been used to parse an input source
 */
const char* xoxa_error_message(ParserState* P)
{
    if (P == NULL) {
        return NULL;
    } else {
        return P->error_message;
    }
}

/**
 * Register a new callback and data glob.
 * @param P a configured parser
 * @param cb a callback function
 * @param data a pointer to arbitrary data, which will be passed to the callback function
 * @return the previous callback.
 */
output_callback_t xoxa_register_output_callback(ParserState* P,
                                                output_callback_t cb,
                                                void* data)
{
    output_callback_t previous = P->callback;
    P->callback = cb;
    P->callback_data = data;
    return previous;
}

// Miscellaneous other helper functions

// Send the string 'str' to the buf, handling whitespace as required.
//
// If skipping_whitespace is NULL, or *skipping_whitespace is
// SKIPPING_WHITESPACE_NONORM, do not skip spaces.
//
// If *skipping_whitespace is SKIPPING_WHITESPACE_COLLAPSE or _TRUE,
// then do skip leading whitespace, with the value of
// *skipping_whitespace being set as appropriate by this function to
// communicate state to the next invocation.
//
// 'len' gives the number of characters to be output.  If 'len == 0',
// then the string is null-terminated.
//
// Return true if any non-whitespace characters were emitted.
static int output_string_with_whitespace_skipping(Buf* buf,
                                                  const XML_Char* str,
                                                  size_t len,
                                                  char *skipping_whitespace)
{
    int found_non_whitespace = 0; // return value
    const byte* start = (const byte*)str;
    const byte* endp;           // end of the text to be output

    if (len == 0) {
        // text is null-terminated
        for (endp=start; *endp!='\0'; endp++) {
            // nothing
        }
    } else {
        endp = start + len;
    }

    // We are processing an XML_Char, which may not, in principle, be
    // a byte.  However if the expat library has not been compiled
    // with the XML_UNICODE parameter, then the input is encoded in
    // UTF-8.  All we care about, in the input string, is whether the
    // characters are whitespace or not; neither the leading or the
    // trailing UTF-8 bytes can be mistaken for whitespace (they all
    // have the high bit set), so we can therefore treat the 'str'
    // contents as bytes.

    if (skipping_whitespace == NULL
        || *skipping_whitespace == SKIPPING_WHITESPACE_NONORM) {
        // this is in the no-normalisation mode
        const byte* p;

        for (p=start; p < endp; p++) {
            switch (*p) {
              case '\n':
              case '\r':
                bufstrcpy(buf, (byte*)(*p == '\n' ? "\\n" : "\\r"));
                break;

              case '\\':
                bufstrcpy(buf, (byte*)"\\\\");
                found_non_whitespace = 1;
                break;

              case 0xc2:
                if (endp - p > 1 && p[1] == 0x85) {
                    // this is U+0085, NEL.  This is not special in
                    // XML 1.0, but is a line-end character in XML
                    // 1.1, so escape it.
                    bufstrcpy(buf, (byte*)"\\u0085");
                    p++;
                } else {
                    found_non_whitespace = 1;
                }
                break;

              case 0xe2:
                if (endp - p > 2 && p[1] == 0x80 && p[2] == 0xa8) {
                    // this is U+2028, which is a line-end character in XML 1.1
                    bufstrcpy(buf, (byte*)"\\u2028");
                    p += 2;
                } else {
                    found_non_whitespace = 1;
                }
                break;

              default:
                found_non_whitespace = 1;
                bufchar(buf, *p);
                break;
            }
        }
    } else {
        // 'words' points to the beginning of an extent between whitespace
        const byte* start = (const byte*)str;
        const byte* words = start;
        const byte* p;
        int last_was_whitespace = 0;

        if (*skipping_whitespace == SKIPPING_WHITESPACE_TRUE) {
            // the immediately preceding invocation of this function left
            // this state value at SKIPPING_WHITESPACE_TRUE, meaning that
            // we're 'currently' skipping whitespace.  So continue to do so.
            while (words < endp) {
                int n = isxmlspace(words);
                if (n) {
                    words += n;
                    last_was_whitespace = 1;
                } else {
                    break;
                }
            }
        }

        p = words;
        while (p < endp) {
            int n = isxmlspace(p);
            if (n) {
                // found whitespace
                last_was_whitespace = 1;
                // copy the 'words' to the buf
                if (p-words > 0) {
                    bufstrncpy(buf, p-words, words);
                }
                // ...jump over this space
                p += n;
                // ...then skip over any remaining space
                while (p < endp) {
                    n = isxmlspace(p);
                    if (n)
                        p += n;
                    else
                        break;
                }
                // ...remember that this is the beginning of a non-whitespace sequence
                words = p;
                // ...and add a single ' ' to the buf
                bufchar(buf, ' ');
            } else {
                found_non_whitespace = 1;
                last_was_whitespace = 0;
                p++;
            }
        }
        if (p-words > 0) {
            // we still have a 'word' left
            bufstrncpy(buf, p-words, words);
        }
        if (last_was_whitespace) {
            // the last character in str was whitespace,
            // which we've skipped, so the next invocation of this
            // function should skip whitespace at the beginning of its
            // buffer, too.
            *skipping_whitespace = SKIPPING_WHITESPACE_TRUE;
        } else {
            // the last character was not whitespace -- any whitespace
            // we encounter when this function is next called should
            // be collapsed rather than skipped.
            *skipping_whitespace = SKIPPING_WHITESPACE_COLLAPSE;
        }
    }

    return found_non_whitespace;
}

/**
 * Returns the version string of the underlying Expat library
 */
const char* xoxa_parser_library_ident(void)
{
    return XML_ExpatVersion();
}
