#ifndef XML_EXPAT_H
#define XML_EXPAT_H 1

#include <stdio.h>
#include <stdint.h>

#include <expat.h>

#include "util_types.h"

struct ParserState_t;

/**
 * The type of the Xoxa callback.
 * The callback takes three arguments:
 * <ul>
 * <li>'bytes' is the array of bytes to process.</li>
 * <li>'len' is the number of bytes in this array to process; len may be passed as zero,
 *     indicating that the byte array is null-terminated.</li>
 * <li>'data' is the pointer registered when the callback is registered
 *     with xoxa_register_output_callback.</li>
 * </ul>
 */
typedef void (*output_callback_t)(const byte* bytes, size_t len, void* data);

output_callback_t xoxa_register_output_callback(struct ParserState_t*, output_callback_t, void*);

// public interface

/**
 * Various flags which control the parser
 * The flags are:
 * <ul>
 *   <li>XOXA_REPORT_ALL  : do not normalise</li>
 *   <li>XOXA_QUIET       : be quiet -- not verbose</li>
 * </ul>
 */
typedef uint8_t ParserFlags;
#define XOXA_REPORT_ALL 1
#define XOXA_QUIET 2

#define INPUT_SOURCE_STRING 0
#define INPUT_SOURCE_FILE 1
typedef struct {
    char type;                  // INPUT_SOURCE_*
    const char* basename;       // the filename, in the case of in.f
    union {
        const char* buffer;
        FILE* f;
    } in;
} InputSource;

// This structure is public, but its contents are not documented
struct ParserState_t {
    XML_Parser parser;
    InputSource *src;
    output_callback_t callback;
    void* callback_data;
    Buf* B;
    Buf* text;
    char text_skipping_whitespace; // values of this are private to xml_expat.c
    ParserFlags parser_flags;
    int signature_fd;
    char* error_message;        // NULL if no error
    XML_StartElementHandler extra_StartElementHandler;
    XML_EndElementHandler extra_EndElementHandler;
    XML_CharacterDataHandler extra_CharacterDataHandler;
    XML_ProcessingInstructionHandler extra_ProcessingInstructionHandler;
    XML_StartCdataSectionHandler extra_StartCdataSectionHandler;
    XML_EndCdataSectionHandler extra_EndCdataSectionHandler;
    XML_StartNamespaceDeclHandler extra_StartNamespaceDeclHandler;
    XML_EndNamespaceDeclHandler extra_EndNamespaceDeclHandler;
    XML_XmlDeclHandler extra_XmlDeclHandler;
    XML_SkippedEntityHandler extra_SkippedEntityHandler;
    void* extra_UserData;
};
typedef struct ParserState_t ParserState;

#if XOXA_PRIVATE_INTERFACE
#define SHOWING_MESSAGES(PS_p) (!((PS_p)->parser_flags & XOXA_QUIET))
#endif /* XOXA_PRIVATE_INTERFACE */

ParserState* xoxa_new_parser(InputSource*);
ParserFlags xoxa_parser_flags(ParserState*, ParserFlags);
void xoxa_release_parser(ParserState*);

/**
 * Parse the contents of the input stream represented by the ParserState argument.
 * Return non-zero on error.
 * If there is an error, then there will be an error message string
 * available through xoxa_error_message(ParserState*).
 */
int xoxa_parse(ParserState*);
void xoxa_signature_fd(ParserState*, int);

/**
 * Return any error message from the ParserState argument.
 * This string must not be freed by any caller.
 */
const char* xoxa_error_message(ParserState*);

InputSource* xoxa_new_input_source_from_string(const char* xmlcontent);
InputSource* xoxa_new_input_source_from_filename(const char* filename);
InputSource* xoxa_new_input_source_from_fd(const int fd);

const char* xoxa_parser_library_ident(void);

#endif /* XML_EXPAT_H */
