#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#include "xoxa.h"

static void Usage(int exit_status);
static const char* progname;


int signature_fd = 0;

void output_callback(const byte* bytes, size_t len, void* data)
{
    FILE* o = (FILE*)data;

    assert(o != NULL);

    if (len == 0) {
        const byte* p;
        for (p=bytes; *p!=0; p++) {
            // nothing
        }
        fwrite(bytes, sizeof(byte), p-bytes, o);
    } else {
        fwrite(bytes, sizeof(byte), len, o);
    }
}

int parse_input(ParserState* PS)
{
    int error_p;

    xoxa_register_output_callback(PS, &output_callback, (void*)stdout);

    error_p = xoxa_parse(PS);
    if (error_p && !(PS->parser_flags & XOXA_QUIET)) {
        fprintf(stderr, "%s: reported a problem: %s\n", progname, xoxa_error_message(PS));
    }
    
    xoxa_release_parser(PS);

    return error_p;
}

static ParserState* create_parser(InputSource* src,
                                  int verbosity,
                                  int report_all_esis,
                                  int signature_fd)
{
    ParserState* PS;
    ParserFlags flags = 0;

    PS = xoxa_new_parser(src);
    if (report_all_esis) {
        flags |= XOXA_REPORT_ALL;
    }
    if (verbosity == 0) {
        flags |= XOXA_QUIET;
    }
    if (signature_fd > 0) {
        xoxa_signature_fd(PS, signature_fd);
    }

    xoxa_parser_flags(PS, flags);

    if (PS == NULL) {
        fprintf(stderr, "Unable to initialise parser state\n");
        exit(1);
    }

    return PS;
}

static int verbosity = 1;

int main(int argc, char** argv) 
{
    enum {
        show_xml,
        show_help,
        show_digest,
    } action = show_xml;

    const char** input_filenames = NULL;
    const char* input_string = NULL;
    int n_inputs = 0;
    int input_no;
    // a MAX_DIGEST_NAME of 16 is surely generous!
#define MAX_DIGEST_NAME 16
    char digest_name[MAX_DIGEST_NAME+1];
    int report_all_esis = 0;
    int exitval;


    progname = argv[0];

    char ch;
    while ((ch = getopt(argc, argv, "ad:hqs:S:V")) != -1) {
        switch (ch) {
          case 'a':
            report_all_esis = 1;
            break;

          case 'd':
            strncpy(digest_name, optarg, MAX_DIGEST_NAME);
            digest_name[MAX_DIGEST_NAME] = '\0';
            action = show_digest;
            break;

          case 's':
            input_string = optarg;
            n_inputs = 1;
            break;

          case 'S':
            {
                int fd;
                char* endp;
                fd = (int)strtol(optarg, &endp, 10);
                if (*endp == '\0') {
                    // valid number
                    signature_fd = fd;
                } else {
                    // deem this to refer to a file, and open this
                    if ((fd = open(optarg, O_WRONLY|O_CREAT, 0666)) < 0) {
                        fprintf(stderr,
                                "Can't open file %s to write\n",
                                optarg);
                        exit(1);
                    }
                    signature_fd = fd;
                }
            }
            break;

          case 'V':
            action = show_help;
            verbosity = 0;
            break;

          case 'q':
            verbosity = 0;
            break;

          case 'h':
            action = show_help;
            verbosity = 1;
            break;

          default:
            Usage(1);
        }
    }
    argc -= optind;
    argv += optind;

    if (argc == 0) {
        input_filenames = NULL;
    } else {
        // there are some arguments
        input_filenames = (const char**)argv;
        n_inputs = argc;
    }

    if (action == show_help) {
        Usage(0);
        /* NOTREACHED */
        assert(0);
    }

    if (input_filenames == NULL) {
        if (input_string == NULL) {
            // no source given on command line, and no filename arguments
            static const char* stdin_input_filenames[] = { "-" };
            input_filenames = stdin_input_filenames;
            n_inputs = 1;
        } else {
            // only an input_string: that's OK
        }
    } else {
        if (input_string == NULL) {
            // only input filenames: that's OK
        } else {
            // input string _and_ input file provided
            Usage(1);
        }
    }
    assert((input_filenames != NULL && n_inputs >= 1)
           || (input_string != NULL && n_inputs == 1));

    exitval = 0;
    for (input_no=0; input_no<n_inputs; input_no++) {
        InputSource* src;
        ParserState* PS;

        if (input_filenames != NULL) {
            if (strcmp(input_filenames[input_no], "-") == 0) {
                // no source given on command line; read from stdin
                src = xoxa_new_input_source_from_fd(0);
            } else {
                src = xoxa_new_input_source_from_filename(input_filenames[input_no]);
            }
            if (src == NULL) {
                fprintf(stderr, "Can't open file %s to read\n", input_filenames[input_no]);
                exit(1);
            }
        } else {
            src = xoxa_new_input_source_from_string(input_string);
            if (src == NULL) {
                fprintf(stderr, "Can't initialise input string\n");
                exit(1);
            }
        }
        assert(src != NULL);
        assert(input_filenames == NULL || input_string == NULL);

        PS = create_parser(src, verbosity, report_all_esis, signature_fd);

        switch (action) {
          case show_xml:
            exitval = parse_input(PS);
            break;

          case show_digest:
              {
                  byte* digest;
                  unsigned int digest_len, i;

                  digest = xoxa_parse_for_digest(PS, digest_name, &digest_len);
                  if (digest) {
                      if (verbosity > 0) {
                          printf("%s(xoxa(%s))= ",
                                 digest_name,
                                 (input_filenames
                                  ? input_filenames[input_no]
                                  : (input_string
                                     ? input_string
                                     : "-")));
                      }
                      for (i=0; i<digest_len; i++) {
                          printf("%02x", digest[i]);
                      }
                      printf("\n");
                      free(digest);
                      exitval = 0;
                  } else {
                      if (verbosity > 0) {
                          const char* msg = xoxa_error_message(PS);
                          fprintf(stderr, "Can't get digest %s (%s)\n",
                                  digest_name,
                                  (msg == NULL ? "<unknown reason>" : msg));
                      }
                      exitval = 1;
                  }

                  xoxa_release_parser(PS);
              }
              break;

          default:
            assert(0);
        }

        if (exitval != 0) {
            break;
        }
    }

    return exitval;
}

static void Usage(int exit_status)
{
    const char* digest_library = xoxa_digest_library_ident();
    FILE* out = (exit_status == 0 ? stdout : stderr);

    fprintf(out, "%s version %s\n", progname, xoxa_version_string());
    fprintf(stderr, "Expat library: %s\n", xoxa_parser_library_ident());
    if (digest_library) {
        fprintf(stderr, "Digest library: %s\n", digest_library);
    }
    if (verbosity) {
        fprintf(out,
                "\n"
                "Usage:\n"
                "    %s [-ahqV] [-ddigest] [input-file ...]\n"
                "    %s [-ahqV] [-ddigest] -s input-string\n"
                "\n"
                "Options:\n"
                "    -a     : show all the output, not just the normalised output\n"
                "    -d digest\n"
                "           : print file digest using named algorithm (eg 'SHA1')\n"
                "    -s str : take input from the given string\n"
                "    -S int|file\n"
                "           : write any found signature to fd or filename\n"
                "    -q     : be quiet (the default is pretty terse)\n"
                "    -h, -V : show help, show version\n"
                "If no source is present, reads from stdin.\n",
                progname, progname);
    }

    exit(exit_status);
}
